#ifdef _WIN32
typedef unsigned char uint8_t;
#else
#include <stdint.h>
#endif

#include <cv.h>
#include <stdio.h>

#include "defs.h"
#include "image.h"
#include "main.h"

#include <math.h>

#include <cv.h>
#include <highgui.h>

#include "geometry.h"

#define RESOLUTION 2
#define THRESHOLD 400

#define TRANS_X ((40*(RESOLUTION))+1)
#define TRANS_Y 1

#define CORR_X 3
#define CORR_Y 3

/* Continuity weight (16 == normal) */
#define X_WEIGHT 16
#define Y_WEIGHT 8

/* Computed macros */
#define THRESH (THRESHOLD*RESOLUTION*CORR_X*CORR_Y)
#define LIMIT_X (TRANS_X+1)
#define WINTRANS_X (CORR_X+TRANS_X-1)
#define WINTRANS_Y (CORR_Y+TRANS_Y-1)

#define WIDTH 4096

typedef struct _window window_t;
struct _window
{
    int r,g,b;
};


typedef struct _pixeldepth pixeldepth_t;
struct _pixeldepth
{
    int score;
    int depth;
    int corr;
    pixeldepth_t *next;
};

static pt2d_t inverse_transform(projection_t *data,
                                pt2d_t from)
{
    pt2d_t to;
    pt3d_t xx, yy, cm;

    xx = multiply(from.x, data->vx);
    yy = multiply(from.y, data->vy);

    cm = add(data->Os, add(xx, yy));

    to.x = cm.x * data->d / cm.z;
    to.y = cm.y * data->d / cm.z;

    return to;
}

static int getpixel_fast_antialiasing(image_t *img, projection_t *data,
                                      double x, double y, int offsetx,
                                      int offsety, int *r, int *g, int *b)
{
    double i0,j0,i1,j1,I_,J_;
    int pR,pB,pG;
    double c1,c2,c3,c4;
    pt2d_t from,to;

    from.x = (float)x / (float)RESOLUTION;
    from.y = y;

    to = inverse_transform(data, from);

    /* Pourcentage de couleur */
    j0 = to.x;
    i0 = to.y;
    j1 = to.x + 1.0;
    i1 = to.y + 1.0;

    J_ = (int)to.x;
    I_ = (int)to.y;

    c1 = ((I_+1) - i0) * ((J_+1) - j0);
    c2 = ((I_+1) - i0) * (j1 - (J_+1));
    c3 = (i1 - (I_+1)) * (j1 - (J_+1));
    c4 = (i1 - (I_+1)) * ((J_+1) - j0);
    pR = pB = pG = 0;
    getpixel_fast(img,(int)to.x+offsetx,(int)to.y+offsety,r,g,b);
    pR = c1 * (*r);
    pG = c1 * (*g);
    pB = c1 * (*b);
    getpixel_fast(img,(int)to.x+1+offsetx,(int)to.y+offsety,r,g,b);
    pR += c2 * (*r);
    pG += c2 * (*g);
    pB += c2 * (*b);
    getpixel_fast(img,(int)to.x+1+offsetx,(int)to.y+1+offsety,r,g,b);
    pR += c3*(*r);
    pG += c3*(*g);
    pB += c3*(*b);
    getpixel_fast(img,(int)to.x+offsetx,(int)to.y+1+offsety,r,g,b);
    pR += c4*(*r);
    pG += c4*(*g);
    pB += c4*(*b);

    *r = pR;
    *g = pG;
    *b = pB;

    return 0;
}

/*****************************************************************************
 * Calcul_bande: calcule la distance entre deux matrices de pixels
 *               de taille CORR_X × CORR_Y
 *****************************************************************************/
static int Calcul_bande(window_t *Refer, window_t *Trans,
                        int dx, int dy, int x, int y,
                        int refer_width, int trans_width)
{
    int i, j;
    int corr = 0;
    int delta;
    window_t P1, P2;

    for(j = -CORR_Y/2 ; j < CORR_Y-CORR_Y/2; j++)
    {
        int k = j * refer_width + x;
        int l = (j + dy) * trans_width + x * RESOLUTION + dx;

        for(i = -CORR_X/2 ; i < CORR_X-CORR_X/2; i++)
        {
            if(i+x < 0 || i+x >= refer_width-1)
            {
                P1.r = 255;
                P1.g = 0;
                P1.b = 255;
            }
            else
            {
                P1 = *(Refer+k+i);
            }

            if(x+i+dx < 0 || i+x*RESOLUTION+dx >= trans_width-1)
            {
                P2.r = 255;
                P2.g = 0;
                P2.b = 255;
            }
            else
            {
                P2 = *(Trans+l+i);
            }

            /* Distance euclidienne */
            delta = sqrt(sq(P1.r-P2.r) + sq(P1.g-P2.g) + sq(P1.b-P2.b));
            corr += delta;
        }
    }

    return corr;
}

/*****************************************************************************
 * Translation_bande: calcule la meilleure corrélation d'une matrice de
 *                    taille CORR_X × CORR_Y dans une fenêtre de taille
 *                    TRANS_X × TRANS_Y.
 *****************************************************************************/
int Translation_bande(int x, int y, window_t *Refer,
                      window_t *Trans, int refer_width,
                      int trans_width, pixeldepth_t *Map, int width)
{
    int dx, dy;
    int corr, corr_min = 0x10000000;
    int dx_min = 0;
    int dold_x = 0, dold_y = 0;
    int good_x = 0, good_y = 0;

    /* Premiere colonne */
    if(x >= 0)
    {
        dold_x = (Map+y*width+x-1)->depth;
        if((Map+y*width+x-1)->corr < THRESH && dold_x >= - LIMIT_X / 2 && dold_x <= LIMIT_X - LIMIT_X / 2)
            good_x = 1;
    }

    /* Premiere ligne */
    if(y >= 2)
    {
        dold_y = (Map+(y-1)*width+x)->depth;
        if((Map+y*width+x-1)->corr < THRESH && dold_y >= - LIMIT_X / 2 && dold_y <= LIMIT_X - LIMIT_X / 2)
            good_y = 1;
    }

    /* Parcours de la translation de gauche a droite */
    for(dy = -TRANS_Y/2; dy < TRANS_Y-TRANS_Y/2; dy++)
    {
        for(dx = -TRANS_X/2; dx < TRANS_X-TRANS_X/2; dx++)
        {
            if(good_x)
            {
                corr = 16 * Calcul_bande(Refer, Trans, dx + dold_x, dy, x, y,
                                         refer_width, trans_width);
                corr += X_WEIGHT * (abs(dx));
                if(good_y)
                    corr += Y_WEIGHT * (abs(dx - (dold_y-dold_x)));
            }
            else
            {
                corr = 16 * Calcul_bande(Refer, Trans, dx, dy, x, y,
                                         refer_width, trans_width);
            }

            /* Minimisation de la corrélation */
            if(corr < corr_min)
            {
                corr_min = corr;
                dx_min = dx;
            }
        }
    }

    (Map+y*width+x)->depth = dx_min + (good_x ? dold_x : 0);
    (Map+y*width+x)->corr = corr_min;
    return corr_min;
}

int Correlation_bande(image_t *left , projection_t *left_norm_data,
                      image_t *right, projection_t *right_norm_data,
                      image_t *left_depthmap_norm,
                      image_t *right_depthmap_norm, double adjust)
{
    int x, y, i;
    double min,max;
    int trans_width, refer_width, map_width, map_height;
    int corr, progress;
    window_t *lines_refer, *lines_trans;
    pixeldepth_t *Map;

    /*Allocation de la structure 'Map' de sortie*/
    map_width = left_norm_data->right - left_norm_data->left;
    map_height = left_norm_data->bottom - left_norm_data->top;
    Map = malloc(sizeof(pixeldepth_t) * map_height * map_width);

    /*Détermination de la taille*/
    //trans_width = (right_norm_data->right-right_norm_data->left)*RESOLUTION + 1;
    //refer_width = left_norm_data->right - left_norm_data->left + 1;
    
    refer_width = WIDTH ;
    trans_width = WIDTH ;

    /*Allocation des lignes*/
    lines_refer = (window_t*)malloc(sizeof(window_t)*refer_width*CORR_Y);
    lines_trans = (window_t*)malloc(sizeof(window_t)*trans_width*WINTRANS_Y);

    
    /*Initialisation des images resultat*/
    cvFillImage(left_depthmap_norm->_private, CV_RGB(0,0,255));
    cvFillImage(right_depthmap_norm->_private, CV_RGB(0,0,255));

    /*Parcours de l'image initiale*/
    for(progress=0,y=0 ; y<map_height; y++)
    {
        while( y * 100 / map_height >= progress)
        {
            printf("\r %i %% effectués", progress);
            fflush(stdout);
            progress ++;
        }

        /* Décalage de ligne */
        for(i = 0; i < CORR_Y - 1; i++ )
            memcpy(lines_refer + refer_width * i,
                   lines_refer + refer_width * (i + 1),
                   sizeof(window_t) * refer_width);

        for(i = 0; i < WINTRANS_Y - 1; i++ )
            memcpy(lines_trans + trans_width * i,
                   lines_trans + trans_width * (i + 1),
                   sizeof(window_t) * trans_width);
        /* Lecture de la ligne suivante de l'image gauche*/
        for(x = 0 ; x < refer_width; x++)
        {
            getpixel_fast_antialiasing(left, left_norm_data, x * RESOLUTION
                                        + (int)left_norm_data->left,
                                       y + (int)left_norm_data->top, 0, 0,
                                       &(lines_refer+x+(CORR_Y-1)*refer_width)->r,
                                       &(lines_refer+x+(CORR_Y-1)*refer_width)->g,
                                       &(lines_refer+x+(CORR_Y-1)*refer_width)->b);
        }
        /* Lecture de la ligne suivante de l'image droite*/
        for(x = 0 ; x < trans_width; x++)
        {
            getpixel_fast_antialiasing(right, right_norm_data,
                                       x + (int)right_norm_data->left,
                                       y + (int)right_norm_data->top, 0, 0,
                                       &(lines_trans+x+(WINTRANS_Y-1)*trans_width)->r,
                                       &(lines_trans+x+(WINTRANS_Y-1)*trans_width)->g,
                                       &(lines_trans+x+(WINTRANS_Y-1)*trans_width)->b);
        }

        /* On attend d'avoir lu WINTRANS_Y lignes */
        if(y < WINTRANS_Y - 1)
            continue;

	/*Parcours de la ligne*/
	for(x = 0; x < map_width; x++)
        {
            /*Correlation*/
#if 1
            corr = Translation_bande(x, y,
                                     lines_refer + refer_width * (CORR_Y/2),
                                     lines_trans + trans_width * (WINTRANS_Y/2),
                                     refer_width, trans_width, Map, map_width);
#else
            corr = 0;
            setpixel_fast(left_depthmap_norm,x,y,
              (lines_trans+l+x)->r,
              (lines_trans+l+x)->g,
              (lines_trans+l+x)->b);
#endif
        }
    }
    printf("100 %% effectués\n");

    /* Normalisation de l'affichage */
    min = 999999; max = -999999;
    for(y=0 ; y<map_height; y++)
    {
        for(x=0 ; x < map_width; x++)
        {
            if((Map+y*map_width+x)->corr < THRESH)
            {
                if((Map+y*map_width+x)->depth < min)
                    min = (Map+y*map_width+x)->depth;

                if((Map+y*map_width+x)->depth > max)
                    max = (Map+y*map_width+x)->depth;
            }
        }
    }
    printf("min/max: %g %g\n",min,max);

    /* Conversion carte de profondeur en image */
    for(y=2 ; y<map_height; y++)
    {
        for(x=0; x<map_width; x++)
        {
            if((Map+y*map_width+x)->corr >= THRESH)
            {
                setpixel_fast(left_depthmap_norm,x,y,255,0,0);
            }
            else
            {
                int pixel = ((float)(max - (Map+y*map_width+x)->depth)
                              / (float)(max - min)) * 255.;
                setpixel(left_depthmap_norm, x, y, pixel, pixel, pixel);
            }
        }
    }

    /* Sauvegarde des images resultats */
    cvSaveImage("sortie.bmp", left_depthmap_norm->_private);
    return 0;
}
