/*
 *  stereo -- Stereo matching software
 *  $Id: limits.c 161 2004-07-02 12:12:32Z sam $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdint.h>
#include <stdio.h>

#include "defs.h"
#include "main.h"
#include "geometry.h"

static pt2d_t transform(projection_t *data,
                        pt2d_t from);

/* Compute destination image limits */
int limits(image_t *image,
           projection_t *data)
{
    int points[][2] =
    {
        { 1, 0 },
        { 0, 1 },
        { 1, 1 }
    };

    pt2d_t from, to;
    int i;

    from.x = from.y = 0.0;
    to = transform(data, from);
    data->left = data->right = to.x;
    data->top = data->bottom = to.y;

    for(i = 0; i < 3; i++)
    {
        from.x = points[i][0] * image->width;
        from.y = points[i][1] * image->height;
        to = transform(data, from);
        if(data->left > to.x)
            data->left = to.x;
        if(data->right < to.x)
            data->right = to.x;
        if(data->top > to.y)
            data->top = to.y;
        if(data->bottom < to.y)
            data->bottom = to.y;
    }

    printf("limites: (%g, %g) (%g, %g)\n", data->left, data->top, data->right, data->bottom);

    return 0;
}

static pt2d_t transform(projection_t *data,
                        pt2d_t from)
{
    pt3d_t pt3d;
    pt2d_t to;

    pt3d.x = from.x;
    pt3d.y = from.y;
    pt3d.z = data->d;

    pt3d = multiply(1 / scalar_product(pt3d, data->Os), pt3d);
    pt3d = multiply(scalar_product(data->Os, data->Os), pt3d);

    to.x = scalar_product(pt3d, data->vx);
    to.y = scalar_product(pt3d, data->vy);

    return to;
}

