/*
 *  stereo -- Stereo matching software
 *  $Id: normalise.c 205 2004-07-30 12:13:49Z eric $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdint.h>
#include <stdio.h>

#include "defs.h"
#include "main.h"
#include "image.h"
#include "geometry.h"

static pt2d_t inverse_transform(projection_t *data,
                                pt2d_t from);

extern int Antialiasing;

/* Normalise an image */
int normalise(image_t *image,
              projection_t *data,
              image_t *dest)
{
    int width, height, x, y, r, g, b;
	double i0,j0,i1,j1,I_,J_;
	int pR,pB,pG;
	double c1,c2,c3,c4;

    width = data->right - data->left;
    height = data->bottom - data->top;

    image_new(dest, width, height, 3);
	if(Antialiasing)
		printf("Antialisasing..\n");
	else
		printf("pas d'antialisasing..\n");
    for(y = data->top; y < data->bottom; y++)
        for(x = data->left; x < data->right; x++)
        {
            pt2d_t to, from;

            from.x = x;
            from.y = y;

            /* XXX: antialiasing */            
            to = inverse_transform(data, from);

    		if(Antialiasing)
    		{
    			/*Pourcentage de couleur */
    			j0 = to.x;
    			i0 = to.y;
    			j1 = to.x + 1.0;
    			i1 = to.y + 1.0;
    			
    			J_ = (int)to.x;
    			I_ = (int)to.y;
    			
    			c1 = ((I_+1) - i0) * ((J_+1) - j0);
    			c2 = ((I_+1) - i0) * (j1 - (J_+1));
    			c3 = (i1 - (I_+1)) * (j1 - (J_+1));
    			c4 = (i1 - (I_+1)) * ((J_+1) - j0);
    			pR = pB = pG = 0;		
    			getpixel(image,(int)to.x,(int)to.y,&r,&g,&b);
    			pR = c1 * r;
    			pG = c1 * g;
    			pB = c1 * b;		
    			getpixel(image,(int)to.x+1,(int)to.y,&r,&g,&b);
    			pR += c2 * r;
    			pG += c2 * g;
    			pB += c2 * b;		
    			getpixel(image,(int)to.x+1,(int)to.y+1,&r,&g,&b);
    			pR += c3*r;
    			pG += c3*g;
    			pB += c3*b;
    			getpixel(image,(int)to.x,(int)to.y+1,&r,&g,&b);
    			pR += c4*r;
    			pG += c4*g;
    			pB += c4*b;		
    	
    			setpixel(dest,(int)from.x-(int)data->left,(int)from.y-(int)data->top,pR,pG,pB);
			}
			else
			{
	            getpixel(image, to.x, to.y, &r, &g, &b);
	            setpixel(dest, x - data->left, y - data->top, r, g, b);
			}
        }

    return 0;
}

static pt2d_t inverse_transform(projection_t *data,
                                pt2d_t from)
{
    pt2d_t to;
    pt3d_t xx, yy, cm;

    xx = multiply(from.x, data->vx);
    yy = multiply(from.y, data->vy);

    cm = add(data->Os, add(xx, yy));

    to.x = cm.x * data->d / cm.z;
    to.y = cm.y * data->d / cm.z;

    return to;
}

